/* It contains some changes needed for libclamav and isn't compatible with
 * the original version, --tk
 */
/* *************************************************************************** {{{
 **
 **  This file is part of the UniquE RAR File Library.
 **
 **  Copyright (C) 2000-2002 by Christian Scheurer (www.ChristianScheurer.ch)
 **  UNIX port copyright (c) 2000-2002 by Johannes Winkelmann (jw@tks6.net)
 **
 **  The contents of this file are subject to the UniquE RAR File Library
 **  License (the "unrarlib-license.txt"). You may not use this file except
 **  in compliance with the License. You may obtain a copy of the License
 **  at http://www.unrarlib.org/license.html.
 **  Software distributed under the License is distributed on an "AS IS"
 **  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied warranty.
 **
 **  Alternatively, the contents of this file may be used under the terms
 **  of the GNU General Public License Version 2 or later (the "GPL"), in
 **  which case the provisions of the GPL are applicable instead of those
 **  above. If you wish to allow use of your version of this file only
 **  under the terms of the GPL and not to allow others to use your version
 **  of this file under the terms of the UniquE RAR File Library License,
 **  indicate your decision by deleting the provisions above and replace
 **  them with the notice and other provisions required by the GPL. If you
 **  do not delete the provisions above, a recipient may use your version
 **  of this file under the terms of the GPL or the UniquE RAR File Library
 **  License.
 ** }}}
 ************************************************************************** */

/* *************************************************************************** {{{
 **
 **                           UniquE RAR File Library
 **                     The free file lib for the demoscene
 **                   multi-OS version (Win32, Linux and SunOS)
 **
 *****************************************************************************
 **
 **   ==> Please configure the program in "unrarlib.h". <==
 **
 **   RAR decompression code:
 **    (C) Eugene Roshal
 **   Modifications to a FileLib:
 **    (C) 2000-2002 Christian Scheurer aka. UniquE/Vantage (cs@unrarlib.org)
 **   Linux port:
 **    (C) 2000-2002 Johannes Winkelmann (jw@tks6.net)
 **
 **  The UniquE RAR File Library gives you the ability to access RAR archives
 **  (any compression method supported in RAR v2.0 including Multimedia
 **  Compression and encryption) directly from your program with ease an by
 **  adding only 12kB (6kB UPX-compressed) additional code to your program.
 **  Both solid and normal (recommended for fast random access to the files!)
 **  archives are supported. This FileLib is made for the Demo scene, so it's
 **  designed for easy use within your demos and intros.
 **  Please read "licence.txt" to learn more about how you may use URARFileLib
 **  in your productions.
 ** }}}
 *****************************************************************************
 **
 **  ==> see the "CHANGES" file to see what's new
 **
 ************************************************************************** */

/* -- include files ------------------------------------------------------- */
#if HAVE_CONFIG_H
#include "clamav-config.h"
#endif

#include "unrarlib.h"                       /* include global configuration */
#include "others.h"
/* ------------------------------------------------------------------------ */

/* -- global stuff -------------------------------------------------------- */
#ifdef _WIN_32	/* WIN32 definitions {{{ */
#include <windows.h>
#include <stdio.h>
#include <string.h>
#define ENABLE_ACCESS
#define HOST_OS     WIN_32
#define FM_NORMAL   0x00
#define FM_RDONLY   0x01
#define FM_HIDDEN   0x02
#define FM_SYSTEM   0x04
#define FM_LABEL    0x08
#define FM_DIREC    0x10
#define FM_ARCH     0x20
#define PATHDIVIDER  "\\"
#define CPATHDIVIDER '\\'
#define MASKALL      "*.*"
#define READBINARY   "rb"
#define READTEXT     "rt"
#define UPDATEBINARY "r+b"
#define CREATEBINARY "w+b"
#define CREATETEXT   "w"
#define APPENDTEXT   "at"
#endif	/* }}} */

#ifdef _UNIX	/* LINUX/UNIX definitions {{{ */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#define ENABLE_ACCESS
#define HOST_OS     UNIX
#define FM_LABEL    0x0000
#define FM_DIREC    0x4000
#define PATHDIVIDER  "/"
#define CPATHDIVIDER '/'
#define MASKALL      "*.*"
#define READBINARY   "r"
#define READTEXT     "r"
#define UPDATEBINARY "r+"
#define CREATEBINARY "w+"
#define CREATETEXT   "w"
#define APPENDTEXT   "a"
/* emulation of the windows API and data types                              */
/* 20-08-2000 Johannes Winkelmann, jw@tks6.net                              */
typedef long    DWORD;
typedef short   BOOL;
#define TRUE    1
#define FALSE   0
#endif	/* }}} */

#ifdef _DEBUG_LOG                           /* define macros for debugging  */
BOOL debug_log_first_start = TRUE;
#define debug_log(a); debug_log_proc(a, __FILE__, __LINE__);
#define debug_init(a); debug_init_proc(a);
void debug_log_proc(char *text, char *sourcefile, int sourceline);
void debug_init_proc(char *file_name);
#else
#define debug_log(a);	cli_dbgmsg("%s:%d %s\n", __FILE__, __LINE__, a);
#define debug_init(a);  /* no debug this time */
#endif


// Struct enum define's {{{


#define MAXWINSIZE      0x100000
#define MAXWINMASK      (MAXWINSIZE-1)
#define UNP_MEMORY      MAXWINSIZE
#define Min(x,y) (((x)<(y)) ? (x):(y))
#define Max(x,y) (((x)>(y)) ? (x):(y))
#define NM  260

#define SIZEOF_MARKHEAD         7
#define SIZEOF_OLDMHD           7
#define SIZEOF_NEWMHD          13
#define SIZEOF_OLDLHD          21
#define SIZEOF_NEWLHD          32
#define SIZEOF_SHORTBLOCKHEAD   7
#define SIZEOF_LONGBLOCKHEAD   11
#define SIZEOF_COMMHEAD        13
#define SIZEOF_PROTECTHEAD     26


#define PACK_VER       20                   /* version of decompression code*/
#define UNP_VER        20
#define PROTECT_VER    20

#define MHD_COMMENT        2
#define MHD_LOCK           4
#define MHD_PACK_COMMENT   16
#define MHD_AV             32
#define MHD_PROTECT        64

#define LHD_SPLIT_BEFORE   1
#define LHD_SPLIT_AFTER    2
#define LHD_PASSWORD       4
#define LHD_COMMENT        8
#define LHD_SOLID          16

#define LHD_WINDOWMASK     0x00e0
#define LHD_WINDOW64       0
#define LHD_WINDOW128      32
#define LHD_WINDOW256      64
#define LHD_WINDOW512      96
#define LHD_WINDOW1024     128
#define LHD_DIRECTORY      0x00e0

#define LONG_BLOCK         0x8000
#define READSUBBLOCK       0x8000

#define NC 298          /* alphabet = {0,1,2, .,NC - 1} */
#define DC 48
#define RC 28
#define BC 19
#define MC 257

#define GetHeaderByte(N) Header[N]
#define GetHeaderWord(N) (Header[N]+((UWORD)Header[N+1]<<8))
#define GetHeaderDword(N) (Header[N]+((UWORD)Header[N+1]<<8)+((UDWORD)Header[N+2]<<16)+((UDWORD)Header[N+3]<<24))
#define AddBits(Bits)                          \
        ctx_rar->InAddr += ( ctx_rar->InBit + (Bits) ) >> 3;     \
        ctx_rar->InBit  =  ( ctx_rar->InBit + (Bits) ) &  7;
#define GetBits()                                                 \
        ctx_rar->BitField = ( ( ( (UDWORD)ctx_rar->InBuf[ctx_rar->InAddr  ] << 16 ) |        \
                      		 ( (UWORD)ctx_rar->InBuf[ctx_rar->InAddr+1] <<  8 ) |        \
                      		 (        ctx_rar->InBuf[ctx_rar->InAddr+2]       ) )        \
                      				 >> (8-ctx_rar->InBit) ) & 0xffff;



enum { M_DENYREAD,M_DENYWRITE,M_DENYNONE,M_DENYALL };
enum { FILE_EMPTY,FILE_ADD,FILE_UPDATE,FILE_COPYOLD,FILE_COPYBLOCK };
enum { SUCCESS,WARNING,FATAL_ERROR,CRC_ERROR,LOCK_ERROR,WRITE_ERROR, OPEN_ERROR,USER_ERROR,MEMORY_ERROR,USER_BREAK=255,IMM_ABORT=0x8000 };
enum { EN_LOCK=1,EN_VOL=2 };
enum { SD_MEMORY=1,SD_FILES=2 };
enum { NAMES_DONTCHANGE };
enum { LOG_ARC=1,LOG_FILE=2 };
enum { OLD_DECODE=0,OLD_ENCODE=1,NEW_CRYPT=2 };
enum { OLD_UNPACK,NEW_UNPACK };
enum { ALL_HEAD=0,MARK_HEAD=0x72,MAIN_HEAD=0x73,FILE_HEAD=0x74, COMM_HEAD=0x75,AV_HEAD=0x76,SUB_HEAD=0x77,PROTECT_HEAD=0x78};
enum { EA_HEAD=0x100 };
enum { MS_DOS=0,OS2=1,WIN_32=2,UNIX=3 };
enum { CODE_HUFFMAN=0,CODE_LZ=1,CODE_LZ2=2,CODE_REPEATLZ=3,CODE_CACHELZ=4, CODE_STARTFILE=5,CODE_ENDFILE=6,CODE_STARTMM=8,CODE_ENDMM=7, 
	CODE_MMDELTA=9};


struct MarkHeader
{
  UBYTE Mark[7];
};


struct NewMainArchiveHeader
{
  UWORD HeadCRC;
  UBYTE HeadType;
  UWORD Flags;
  UWORD HeadSize;
  UWORD Reserved;
  UDWORD Reserved1;
};


struct NewFileHeader
{
  UWORD HeadCRC;
  UBYTE HeadType;
  UWORD Flags;
  UWORD HeadSize;
  UDWORD PackSize;
  UDWORD UnpSize;
  UBYTE HostOS;
  UDWORD FileCRC;
  UDWORD FileTime;
  UBYTE UnpVer;
  UBYTE Method;
  UWORD NameSize;
  UDWORD FileAttr;
};


struct BlockHeader
{
  UWORD HeadCRC;
  UBYTE HeadType;
  UWORD Flags;
  UWORD HeadSize;
  UDWORD DataSize;
};


struct Decode
{
  unsigned int MaxNum;
  unsigned int DecodeLen[16];
  unsigned int DecodePos[16];
  unsigned int DecodeNum[2];
};


struct AudioVariables
{
  int K1,K2,K3,K4,K5;
  int D1,D2,D3,D4;
  int LastDelta;
  unsigned int Dif[11];
  unsigned int ByteCount;
  int LastChar;
};

struct LitDecode
{
  unsigned int MaxNum;
  unsigned int DecodeLen[16];
  unsigned int DecodePos[16];
  unsigned int DecodeNum[NC];
};

struct DistDecode
{
  unsigned int MaxNum;
  unsigned int DecodeLen[16];
  unsigned int DecodePos[16];
  unsigned int DecodeNum[DC];
};

struct RepDecode
{
  unsigned int MaxNum;
  unsigned int DecodeLen[16];
  unsigned int DecodePos[16];
  unsigned int DecodeNum[RC];
};

struct MultDecode
{
  unsigned int MaxNum;
  unsigned int DecodeLen[16];
  unsigned int DecodePos[16];
  unsigned int DecodeNum[MC];
};

struct BitDecode
{
  unsigned int MaxNum;
  unsigned int DecodeLen[16];
  unsigned int DecodePos[16];
  unsigned int DecodeNum[BC];
};

// }}}

struct sRAR {
	int	is_init;
	UDWORD Key[4];					// OK
	UBYTE SubstTable[256];				// OK
	int UnpAudioBlock;				// OK
	int UnpChannels;				// OK
	int CurChannel;					// OK
	int ChannelDelta;				// OK
	unsigned int OldDist[4];			// OK
	unsigned int OldDistPtr;			// OK
	unsigned int Length;				// OK
	unsigned int Distance;				// OK
	unsigned int LastDist;				// OK
	unsigned int LastLength;			// OK
	unsigned int InAddr;				// OK
	unsigned int InBit;				// OK
	unsigned int ReadTop;				// OK
	unsigned int UnpPtr;				// OK
	unsigned int WrPtr;				// OK
	int Encryption;					// OK
	UDWORD HeaderCRC;				// OK
	long UnpPackedSize;				// OK
	long DestUnpSize;				// OK
	unsigned long CurUnpRead;			// OK
	unsigned long CurUnpWrite;			// OK
	long CurBlockPos;				// OK
	long NextBlockPos;				// OK
	int MainHeadSize;				// OK
	unsigned char *temp_output_buffer;          	//OK /* extract files to this pointer*/
	unsigned long *temp_output_buffer_offset;   	//OK /* size of temp. extract buffer */
	char Password[255];				//OK /* password to decrypt files    */
	UBYTE *UnpMemory;
	UBYTE *TempMemory; 				/* temporary unpack-buffer      */
	char  *CommMemory;
	char ArgName[NM];				//OK /* current file in rar archive  */
	char ArcFileName[NM];				//OK /* file to decompress           */
	MemoryFile *MemRARFile;				//OK /* mem-to-mem decompression pointer to RAR file in memory*/
	FILE *ArcPtr;					//OK /* input RAR file handler       */
	struct MarkHeader MarkHead;			//OK
	struct NewMainArchiveHeader NewMhd;		//OK
	struct NewFileHeader NewLhd;			//OK
	struct BlockHeader BlockHead;			//OK
	struct AudioVariables AudV[4];			//OK
	unsigned char *UnpBuf;				//OK
	unsigned int   BitField;			//OK
	unsigned int   Number;				//OK
	unsigned char  InBuf[8192];			//OK /* input read buffer            */
	unsigned char  UnpOldTable[MC*4];		//OK
	struct LitDecode   LD;				//OK
	struct DistDecode  DD;				//OK
	struct RepDecode   RD;				//OK
	struct BitDecode   BD;				//OK
	struct MultDecode  MD[4];			//OK
	struct MultDecode *MDPtr[4];			//OK
	int 		   LastBlock;
	};

static struct sRAR  ctx_rar_buf;

// {{{ String functions 
static char* strupper(char *Str) {{{
  	char *ChPtr;
  	for (ChPtr=Str;*ChPtr;ChPtr++) *ChPtr=(char)toupper(*ChPtr);
  	return(Str);
	}}}
static int stricomp(char *Str1,char *Str2) {{{
/* compare strings without regard of '\' and '/'                            
 * Old Version was not thread safe NOR correct since '\\' == '/' == '_'
 * */
  	while(*Str1 && *Str2 && (*Str1==*Str2 || (*Str1=='/' && *Str2=='\\') || (*Str1=='\\' && *Str2=='/'))) {
		Str1++;
		Str2++;
  		}
	if (*Str1 <  *Str2) return -1;
	if (*Str1 >  *Str2) return  1;
	if (*Str1 == *Str2) return  0;
//  return(strcmp(strupper(S1),strupper(S2)));
}}}
// }}}

static void RAR_init(struct sRAR *ctx_rar) {{{
	if (ctx_rar->is_init) return;
	ctx_rar->UnpMemory  = NULL;
	ctx_rar->TempMemory = NULL;
	ctx_rar->CommMemory = NULL;
	ctx_rar->MDPtr[0] = &ctx_rar->MD[0];
	ctx_rar->MDPtr[1] = &ctx_rar->MD[1];
	ctx_rar->MDPtr[2] = &ctx_rar->MD[2];
	ctx_rar->MDPtr[3] = &ctx_rar->MD[3];
	ctx_rar->is_init = 1;
	}}}

/* *************************************************** 
 * ** CRCCrypt Code - decryption engine starts here ** {{{
 * ***************************************************/
#define NROUNDS 32
#define rol(x,n)  (((x)<<(n)) | ((x)>>(8*sizeof(x)-(n))))
#define ror(x,n)  (((x)>>(n)) | ((x)<<(8*sizeof(x)-(n))))
#define substLong(t) ( (UDWORD)SubstTable[(int)t&255] | \
           ((UDWORD)SubstTable[(int)(t>> 8)&255]<< 8) | \
           ((UDWORD)SubstTable[(int)(t>>16)&255]<<16) | \
           ((UDWORD)SubstTable[(int)(t>>24)&255]<<24) )
static const UBYTE InitSubstTable[256]={
  215, 19,149, 35, 73,197,192,205,249, 28, 16,119, 48,221,  2, 42,
  232,  1,177,233, 14, 88,219, 25,223,195,244, 90, 87,239,153,137,
  255,199,147, 70, 92, 66,246, 13,216, 40, 62, 29,217,230, 86,  6,
   71, 24,171,196,101,113,218,123, 93, 91,163,178,202, 67, 44,235,
  107,250, 75,234, 49,167,125,211, 83,114,157,144, 32,193,143, 36,
  158,124,247,187, 89,214,141, 47,121,228, 61,130,213,194,174,251,
   97,110, 54,229,115, 57,152, 94,105,243,212, 55,209,245, 63, 11,
  164,200, 31,156, 81,176,227, 21, 76, 99,139,188,127, 17,248, 51,
  207,120,189,210,  8,226, 41, 72,183,203,135,165,166, 60, 98,  7,
  122, 38,155,170, 69,172,252,238, 39,134, 59,128,236, 27,240, 80,
  131,  3, 85,206,145, 79,154,142,159,220,201,133, 74, 64, 20,129,
  224,185,138,103,173,182, 43, 34,254, 82,198,151,231,180, 58, 10,
  118, 26,102, 12, 50,132, 22,191,136,111,162,179, 45,  4,148,108,
  161, 56, 78,126,242,222, 15,175,146, 23, 33,241,181,190, 77,225,
    0, 46,169,186, 68, 95,237, 65, 53,208,253,168,  9, 18,100, 52,
  116,184,160, 96,109, 37, 30,106,140,104,150,  5,204,117,112, 84
};

// }}}


// Thread Safe CRC32 Functions {{{
static UDWORD CRCTab[256];
static short crcInitialized = 0;
void InitCRC(void) {{{
  	int I, J;
  	UDWORD C;
  	if (crcInitialized) return;
  	cli_dbgmsg("%s:%d:%s Initialize CRC table\n", __FILE__, __LINE__, "InitCRC");
  	for (I=0;I<256;I++) {
    		for (C=I,J=0;J<8;J++) C=(C & 1) ? (C>>1)^0xEDB88320L : (C>>1);
    		CRCTab[I]=C;
  		}
  	crcInitialized = 1;
	}}}
static UDWORD CalcCRC32(UDWORD StartCRC,UBYTE *Addr,UDWORD Size) {{{
  	unsigned int I;
  	for (I=0; I<Size; I++) StartCRC = CRCTab[(UBYTE)StartCRC ^ Addr[I]] ^ (StartCRC >> 8);
  	return(StartCRC);
	}}}
// }}}

static void         tseek       (struct sRAR *ctx_rar, long offset,int fromwhere) {{{
#ifndef _USE_MEMORY_TO_MEMORY_DECOMPRESSION	
	FILE *stream = ctx_rar->ArcPtr;
	if (ctx_rar->ArcPtr == NULL) return;
	fseek(stream,offset,fromwhere);
#else
	if(fromwhere==SEEK_SET) ctx_rar->MemRARFile->offset  = ctx_rar->NextBlockPos;
	if(fromwhere==SEEK_CUR) ctx_rar->MemRARFile->offset += ctx_rar->NextBlockPos;
#endif
	}}}
static int          tread       (struct sRAR *ctx_rar, void *buf,unsigned len) {{{
#ifdef _USE_MEMORY_TO_MEMORY_DECOMPRESSION
	void *stream = ctx_rar->MemRARFile;
	if(((ctx_rar->MemRARFile->offset + len) > ctx_rar->MemRARFile->size) || (len == 0)) return 0;
	memcpy(buf, (BYTE*)(((MemoryFile*)stream)->data)+((MemoryFile*)stream)->offset, len % ((((MemoryFile*)stream)->size) - 1));
	ctx_rar->MemRARFile->offset+=len;                  /* update read pointer          */
	return len % ((((MemoryFile*)stream)->size) - 1);
#else
	FILE *stream = ctx_rar->ArcPtr;
	return(fread(buf,1,len,(FILE*)stream));
#endif
	}}}
static long         tftell      (struct sRAR *ctx_rar) {{{
#ifdef _USE_MEMORY_TO_MEMORY_DECOMPRESSION
      	return ctx_rar->MemRARFile->offset;       /* get offset of mem-file       */
#else
      	return ftell(ctx_rar->ArcPtr);
#endif
	}}}

static int          ReadHeader  (struct sRAR *ctx_rar, int BlockType) {{{
  int Size = 0;
  unsigned char Header[64];
  memset(Header, 0, sizeof(Header));
  switch(BlockType)
  {
    case MAIN_HEAD:
        Size=tread(ctx_rar, Header, SIZEOF_NEWMHD);
        ctx_rar->NewMhd.HeadCRC=(unsigned short)GetHeaderWord(0);
        ctx_rar->NewMhd.HeadType=GetHeaderByte(2);
        ctx_rar->NewMhd.Flags=(unsigned short)GetHeaderWord(3);
        ctx_rar->NewMhd.HeadSize=(unsigned short)GetHeaderWord(5);
        ctx_rar->NewMhd.Reserved=(unsigned short)GetHeaderWord(7);
        ctx_rar->NewMhd.Reserved1=GetHeaderDword(9);
        ctx_rar->HeaderCRC=CalcCRC32(0xFFFFFFFFL,&Header[2],SIZEOF_NEWMHD-2);
      break;
    case FILE_HEAD:
        Size=tread(ctx_rar, Header, SIZEOF_NEWLHD);
        ctx_rar->NewLhd.HeadCRC=(unsigned short)GetHeaderWord(0);
        ctx_rar->NewLhd.HeadType=GetHeaderByte(2);
        ctx_rar->NewLhd.Flags=(unsigned short)GetHeaderWord(3);
        ctx_rar->NewLhd.HeadSize=(unsigned short)GetHeaderWord(5);
        ctx_rar->NewLhd.PackSize=GetHeaderDword(7);
        ctx_rar->NewLhd.UnpSize=GetHeaderDword(11);
        ctx_rar->NewLhd.HostOS=GetHeaderByte(15);
        ctx_rar->NewLhd.FileCRC=GetHeaderDword(16);
        ctx_rar->NewLhd.FileTime=GetHeaderDword(20);
        ctx_rar->NewLhd.UnpVer=GetHeaderByte(24);
        ctx_rar->NewLhd.Method=GetHeaderByte(25);
        ctx_rar->NewLhd.NameSize=(unsigned short)GetHeaderWord(26);
        ctx_rar->NewLhd.FileAttr=GetHeaderDword(28);
        ctx_rar->HeaderCRC=CalcCRC32(0xFFFFFFFFL,Header+2,SIZEOF_NEWLHD-2);
      break;

#ifdef _DEBUG_LOG
  case COMM_HEAD:                           /* log errors in case of debug  */
        debug_log("Comment headers not supported! "\
                  "Please create archives without comments.");
      break;
  case PROTECT_HEAD:
        debug_log("Protected headers not supported!");
      break;
  case ALL_HEAD:
        debug_log("ShortBlockHeader not supported!");
      break;
  default:
        debug_log("Unknown//unsupported !");
#else
  default:                                  /* else do nothing              */
        break;
#endif
  }
  return(Size);
}}}
static int          ReadBlock   (struct sRAR *ctx_rar, int BlockType) {{{
  	struct NewFileHeader SaveFileHead;
  	int Size=0,ReadSubBlock=0;
  	memcpy(&SaveFileHead,&ctx_rar->NewLhd,sizeof(SaveFileHead));
  	if (BlockType & READSUBBLOCK) {
    		ReadSubBlock=1;
    		BlockType &= 0xff;
  		}
    	while (1) {
	    	ctx_rar->CurBlockPos=tftell (ctx_rar);
      		Size=ReadHeader(ctx_rar, FILE_HEAD);
      		if (Size!=0) {
        		if (ctx_rar->NewLhd.HeadSize<SIZEOF_SHORTBLOCKHEAD) return(0);
        		ctx_rar->NextBlockPos=ctx_rar->CurBlockPos+ctx_rar->NewLhd.HeadSize;
        		if (ctx_rar->NewLhd.Flags & LONG_BLOCK) ctx_rar->NextBlockPos+=ctx_rar->NewLhd.PackSize;
        		if (ctx_rar->NextBlockPos<=ctx_rar->CurBlockPos) return(0);
      			}
      		if (Size > 0 && BlockType!=SUB_HEAD) ctx_rar->LastBlock=BlockType;
      		if (Size==0 || BlockType==ALL_HEAD || ctx_rar->NewLhd.HeadType==BlockType || 
				(ctx_rar->NewLhd.HeadType==SUB_HEAD && ReadSubBlock && ctx_rar->LastBlock==BlockType))
        		break;
      		tseek(ctx_rar, ctx_rar->NextBlockPos, SEEK_SET);
    		}
  	ctx_rar->BlockHead.HeadCRC	=ctx_rar->NewLhd.HeadCRC;
  	ctx_rar->BlockHead.HeadType	=ctx_rar->NewLhd.HeadType;
  	ctx_rar->BlockHead.Flags	=ctx_rar->NewLhd.Flags;
  	ctx_rar->BlockHead.HeadSize	=ctx_rar->NewLhd.HeadSize;
  	ctx_rar->BlockHead.DataSize	=ctx_rar->NewLhd.PackSize;

  	if (BlockType!=ctx_rar->NewLhd.HeadType) BlockType=ALL_HEAD;
  	if((FILE_HEAD == BlockType) && (Size>0)) {
    		ctx_rar->NewLhd.NameSize=Min(ctx_rar->NewLhd.NameSize,sizeof(ctx_rar->ArcFileName)-1);
    		tread(ctx_rar, ctx_rar->ArcFileName, ctx_rar->NewLhd.NameSize);
    		ctx_rar->ArcFileName[ctx_rar->NewLhd.NameSize]=0;
#ifdef _DEBUG_LOG
    		if (ctx_rar->NewLhd.HeadCRC!=(UWORD)~CalcCRC32(ctx_rar->HeaderCRC,(UBYTE*)&ctx_rar->ArcFileName[0], ctx_rar->NewLhd.NameSize)) {
      			debug_log("file header broken");
		}
#endif
    		Size+=ctx_rar->NewLhd.NameSize;
  		} 
	else {
    		memcpy(&ctx_rar->NewLhd,&SaveFileHead,sizeof(ctx_rar->NewLhd));
    		tseek(ctx_rar, ctx_rar->CurBlockPos,SEEK_SET);
  		}
  return(Size);
}}}
static void         UpdKeys     (struct sRAR *ctx_rar, UBYTE *Buf) {{{
  int I;
  UDWORD *Key = ctx_rar->Key;
  for (I=0;I<16;I+=4)
  {
    Key[0]^=CRCTab[Buf[I  ]];
    Key[1]^=CRCTab[Buf[I+1]];
    Key[2]^=CRCTab[Buf[I+2]];
    Key[3]^=CRCTab[Buf[I+3]];
  }
}}}
static void         DecryptBlock(struct sRAR *ctx_rar, UBYTE *Buf) {{{
  int I;
  UBYTE InBuf[16];
  UDWORD A,B,C,D,T,TA,TB;
  UDWORD *Key = ctx_rar->Key;
  UBYTE *SubstTable = ctx_rar->SubstTable;
#ifdef NON_INTEL_BYTE_ORDER
  A=((UDWORD)Buf[ 0]|((UDWORD)Buf[ 1]<<8)|((UDWORD)Buf[ 2]<<16)| ((UDWORD)Buf[ 3]<<24))^Key[0];
  B=((UDWORD)Buf[ 4]|((UDWORD)Buf[ 5]<<8)|((UDWORD)Buf[ 6]<<16)| ((UDWORD)Buf[ 7]<<24))^Key[1];
  C=((UDWORD)Buf[ 8]|((UDWORD)Buf[ 9]<<8)|((UDWORD)Buf[10]<<16)| ((UDWORD)Buf[11]<<24))^Key[2];
  D=((UDWORD)Buf[12]|((UDWORD)Buf[13]<<8)|((UDWORD)Buf[14]<<16)| ((UDWORD)Buf[15]<<24))^Key[3];
#else
  UDWORD *BufPtr;
  BufPtr=(UDWORD *)Buf;
  A=BufPtr[0]^Key[0];                       /* xxx may be this can be       */
  B=BufPtr[1]^Key[1];                       /* optimized in assembler       */
  C=BufPtr[2]^Key[2];
  D=BufPtr[3]^Key[3];
#endif
  memcpy(ctx_rar->InBuf,Buf,sizeof(ctx_rar->InBuf));
  for(I=NROUNDS-1;I>=0;I--)
  {
    T=((C+rol(D,11))^Key[I&3]);
    TA=A^substLong(T);
    T=((D^rol(C,17))+Key[I&3]);
    TB=B^substLong(T);
    A=C;
    B=D;
    C=TA;
    D=TB;
  }
#ifdef NON_INTEL_BYTE_ORDER
  C^=Key[0];
  Buf[0]=(UBYTE)C;
  Buf[1]=(UBYTE)(C>>8);
  Buf[2]=(UBYTE)(C>>16);
  Buf[3]=(UBYTE)(C>>24);
  D^=Key[1];
  Buf[4]=(UBYTE)D;
  Buf[5]=(UBYTE)(D>>8);
  Buf[6]=(UBYTE)(D>>16);
  Buf[7]=(UBYTE)(D>>24);
  A^=Key[2];
  Buf[8]=(UBYTE)A;
  Buf[9]=(UBYTE)(A>>8);
  Buf[10]=(UBYTE)(A>>16);
  Buf[11]=(UBYTE)(A>>24);
  B^=Key[3];
  Buf[12]=(UBYTE)B;
  Buf[13]=(UBYTE)(B>>8);
  Buf[14]=(UBYTE)(B>>16);
  Buf[15]=(UBYTE)(B>>24);
#else
  BufPtr[0]=C^Key[0];
  BufPtr[1]=D^Key[1];
  BufPtr[2]=A^Key[2];
  BufPtr[3]=B^Key[3];
#endif
  UpdKeys(ctx_rar, ctx_rar->InBuf);
}}}
static unsigned int UnpRead     (struct sRAR *ctx_rar, unsigned char *Addr,unsigned int Count) {{{
  	int RetCode=0;
  	unsigned int I,ReadSize,TotalRead=0;
  	unsigned char *ReadAddr;
  	ReadAddr=Addr;
  	while (Count > 0) {
    		ReadSize=(unsigned int)((Count>(unsigned long)ctx_rar->UnpPackedSize) ?  ctx_rar->UnpPackedSize : Count);
#ifdef _USE_MEMORY_TO_MEMORY_DECOMPRESSION
    		cli_dbgmsg("unrarlib: UnpREAD: Using memory->memory decompression\n");
    		if(ctx_rar->MemRARFile->data == NULL) return(0);
#else
    		if (ctx_rar->ArcPtr==NULL) return(0);
#endif
    		RetCode=tread(ctx_rar, ReadAddr, ReadSize);
    		ctx_rar->CurUnpRead+=RetCode;
    		ReadAddr+=RetCode;
    		TotalRead+=RetCode;
    		Count-=RetCode;
    		ctx_rar->UnpPackedSize-=RetCode;
      		break;
  		}
  cli_dbgmsg("ctx_rar->CurUnpRead == %d, TotalRead == %d, Count == %d, ctx_rar->UnpPackedSize == %d\n", ctx_rar->CurUnpRead, TotalRead, Count, ctx_rar->UnpPackedSize);
  if (RetCode!= -1) {
    RetCode=TotalRead;
    if (ctx_rar->Encryption) {
      if (ctx_rar->Encryption<20) {
            cli_dbgmsg("unrarlib: Old Crypt() not supported!");
          }
      else {
        for (I=0;I<(unsigned int)RetCode;I+=16) DecryptBlock(ctx_rar, &Addr[I]);
      }
    }
  }
  return(RetCode);
}}}
static int          IsArchive   (struct sRAR *ctx_rar) {{{
#ifdef _DEBUG_LOG
  int  str_offs;                            /* used for debug-strings       */
  char DebugMsg[500];                       /* used to compose debug msg    */
#endif

  if (tread(ctx_rar, ctx_rar->MarkHead.Mark, SIZEOF_MARKHEAD) != SIZEOF_MARKHEAD) {
    debug_log("IsArchive(): short read: FALSE");
    return(FALSE);
  }
  /* Old archive => error                                                   */
  if (ctx_rar->MarkHead.Mark[0]==0x52 && ctx_rar->MarkHead.Mark[1]==0x45 &&
      ctx_rar->MarkHead.Mark[2]==0x7e && ctx_rar->MarkHead.Mark[3]==0x5e)
  {
    debug_log("Attention: format as OLD detected! Can't handle archive!");
  }
  else
      /* original RAR v2.0                                                  */
      if ((ctx_rar->MarkHead.Mark[0]==0x52 && ctx_rar->MarkHead.Mark[1]==0x61 && /* original  */
           ctx_rar->MarkHead.Mark[2]==0x72 && ctx_rar->MarkHead.Mark[3]==0x21 && /* RAR header*/
           ctx_rar->MarkHead.Mark[4]==0x1a && ctx_rar->MarkHead.Mark[5]==0x07 &&
           ctx_rar->MarkHead.Mark[6]==0x00) ||
     /* "UniquE!" - header                                                  */
          (ctx_rar->MarkHead.Mark[0]=='U' && ctx_rar->MarkHead.Mark[1]=='n' &&   /* "UniquE!" */
           ctx_rar->MarkHead.Mark[2]=='i' && ctx_rar->MarkHead.Mark[3]=='q' &&   /* header    */
           ctx_rar->MarkHead.Mark[4]=='u' && ctx_rar->MarkHead.Mark[5]=='E' &&
           ctx_rar->MarkHead.Mark[6]=='!'))

    {
      if (ReadHeader(ctx_rar, MAIN_HEAD)!=SIZEOF_NEWMHD) {
        debug_log("IsArchive(): ReadHeader() failed");
        return(FALSE);
      }
    } else
    {

#ifdef _DEBUG_LOG
     /* sorry for this ugly code, but older SunOS gcc compilers don't       */
     /* support white spaces within strings                                 */
     str_offs  = sprintf(DebugMsg, "unknown archive type (only plain RAR ");
     str_offs += sprintf(DebugMsg + str_offs, "supported (normal and solid ");
     str_offs += sprintf(DebugMsg + str_offs, "archives), SFX and Volumes ");
     str_offs += sprintf(DebugMsg + str_offs, "are NOT supported!)");

     debug_log(DebugMsg);
#endif

    }


  ctx_rar->MainHeadSize=SIZEOF_NEWMHD;

  return(TRUE);
}}}
static BOOL         UnstoreFile (struct sRAR *ctx_rar) {{{
  if ((long)(*ctx_rar->temp_output_buffer_offset=UnpRead(ctx_rar, ctx_rar->temp_output_buffer, ctx_rar->NewLhd.UnpSize))==-1)
  {
    cli_dbgmsg("unrarlib: Read error of stored file!");
    return FALSE;
  }
  return TRUE;
}}}
static void         EncryptBlock(struct sRAR *ctx_rar, UBYTE *Buf) {{{
  int I;
  UDWORD *Key = ctx_rar->Key;
  UBYTE *SubstTable = ctx_rar->SubstTable;
  UDWORD A,B,C,D,T,TA,TB;
#ifdef NON_INTEL_BYTE_ORDER
  A=((UDWORD)Buf[ 0]|((UDWORD)Buf[ 1]<<8)|((UDWORD)Buf[ 2]<<16)| ((UDWORD)Buf[ 3]<<24))^Key[0];
  B=((UDWORD)Buf[ 4]|((UDWORD)Buf[ 5]<<8)|((UDWORD)Buf[ 6]<<16)| ((UDWORD)Buf[ 7]<<24))^Key[1];
  C=((UDWORD)Buf[ 8]|((UDWORD)Buf[ 9]<<8)|((UDWORD)Buf[10]<<16)| ((UDWORD)Buf[11]<<24))^Key[2];
  D=((UDWORD)Buf[12]|((UDWORD)Buf[13]<<8)|((UDWORD)Buf[14]<<16)| ((UDWORD)Buf[15]<<24))^Key[3];
#else
  UDWORD *BufPtr;
  BufPtr=(UDWORD *)Buf;
  A=BufPtr[0]^Key[0];
  B=BufPtr[1]^Key[1];
  C=BufPtr[2]^Key[2];
  D=BufPtr[3]^Key[3];
#endif
  for(I=0;I<NROUNDS;I++)
  {
    T=((C+rol(D,11))^Key[I&3]);
    TA=A^substLong(T);
    T=((D^rol(C,17))+Key[I&3]);
    TB=B^substLong(T);
    A=C;
    B=D;
    C=TA;
    D=TB;
  }
#ifdef NON_INTEL_BYTE_ORDER
  C^=Key[0];
  Buf[0]=(UBYTE)C;
  Buf[1]=(UBYTE)(C>>8);
  Buf[2]=(UBYTE)(C>>16);
  Buf[3]=(UBYTE)(C>>24);
  D^=Key[1];
  Buf[4]=(UBYTE)D;
  Buf[5]=(UBYTE)(D>>8);
  Buf[6]=(UBYTE)(D>>16);
  Buf[7]=(UBYTE)(D>>24);
  A^=Key[2];
  Buf[8]=(UBYTE)A;
  Buf[9]=(UBYTE)(A>>8);
  Buf[10]=(UBYTE)(A>>16);
  Buf[11]=(UBYTE)(A>>24);
  B^=Key[3];
  Buf[12]=(UBYTE)B;
  Buf[13]=(UBYTE)(B>>8);
  Buf[14]=(UBYTE)(B>>16);
  Buf[15]=(UBYTE)(B>>24);
#else
  BufPtr[0]=C^Key[0];
  BufPtr[1]=D^Key[1];
  BufPtr[2]=A^Key[2];
  BufPtr[3]=B^Key[3];
#endif
  UpdKeys(ctx_rar, Buf);
}}}
static void         SetCryptKeys(struct sRAR *ctx_rar, char *NewPassword) {{{
  unsigned int I,J,K,PswLength;
  unsigned char N1,N2;
  unsigned char Psw[256];
  UDWORD *Key = ctx_rar->Key;
  UBYTE *SubstTable = ctx_rar->SubstTable;

#if !defined _USE_ASM
  UBYTE Ch;
#endif

//  SetOldKeys(NewPassword);

  Key[0]=0xD3A3B879L;
  Key[1]=0x3F6D12F7L;
  Key[2]=0x7515A235L;
  Key[3]=0xA4E7F123L;
  memset(Psw,0,sizeof(Psw));
  strcpy((char *)Psw,NewPassword);
  PswLength=strlen(NewPassword);
  memcpy(SubstTable,InitSubstTable,sizeof(SubstTable));

  for (J=0;J<256;J++)
    for (I=0;I<PswLength;I+=2)
    {
      N2=(unsigned char)CRCTab[(Psw[I+1]+J)&0xFF];
      for (K=1, N1=(unsigned char)CRCTab[(Psw[I]-J)&0xFF];
           (N1!=N2);
           N1++, K++)
          {
#ifdef _USE_ASM

#ifdef _WIN_32
          __asm {

                    mov ebx, Offset SubstTable
                    mov edx, ebx

                    xor ecx, ecx            /* read SubstTable[N1]... */
                    mov cl, N1
                    add ebx, ecx
                    mov al, byte ptr[ebx]

                    mov cl, N1              /* read SubstTable[(N1+I+K)&0xFF]... */
                    add ecx, I
                    add ecx, K
                    and ecx, 0xFF
                    add edx, ecx
                    mov ah, byte ptr[edx]

                    mov  byte ptr[ebx], ah  /* and write back */
                    mov  byte ptr[edx], al

               }
#else
                     __asm__ __volatile__ (
"                    xorl %%ecx, %%ecx"
"                    movl %2, %%ecx                     /* ecx = N1 */"
"                    mov %%ebx, %%edx"
"                    addl %%ecx, %%ebx"
""
"                    addl %0, %%ecx"
"                    addl %1, %%ecx"
"                    andl $0x000000FF, %%ecx"
"                    addl %%ecx, %%edx"
"                    "
"                    movb (%%ebx), %%al"
"                    movb (%%edx), %%ah"
""
"                    movb  %%ah, (%%ebx)     /* and write back */"
"                    movb  %%al, (%%edx)"
                    : : "g" ((long)I),
                        "g" ((long)K),
                        "g" ((long)N1),
                        "ebx"((long)SubstTable)
                    : "ecx", "edx"

              );
#endif

#else
            /* Swap(&SubstTable[N1],&SubstTable[(N1+I+K)&0xFF]);            */
             Ch=SubstTable[N1];
             SubstTable[N1]=SubstTable[(N1+I+K)&0xFF];
             SubstTable[(N1+I+K)&0xFF]=Ch;
#endif
          }
    }
  for (I=0;I<PswLength;I+=16)
    EncryptBlock(ctx_rar, &Psw[I]);
}}}
static UBYTE        DecodeAudio (struct sRAR *ctx_rar, int Delta) {{{
  struct AudioVariables *V;
  unsigned int Ch;
  unsigned int NumMinDif,MinDif;
  int PCh,I;

  V=&ctx_rar->AudV[ctx_rar->CurChannel];
  V->ByteCount++;
  V->D4=V->D3;
  V->D3=V->D2;
  V->D2=V->LastDelta-V->D1;
  V->D1=V->LastDelta;
  PCh=8*V->LastChar+V->K1*V->D1+V->K2*V->D2+
           V->K3*V->D3+V->K4*V->D4+V->K5*ctx_rar->ChannelDelta;
  PCh=(PCh>>3) & 0xFF;

  Ch=PCh-Delta;

  I=((signed char)Delta)<<3;

  V->Dif[ 0]+=abs(I);
  V->Dif[ 1]+=abs(I-V->D1);
  V->Dif[ 2]+=abs(I+V->D1);
  V->Dif[ 3]+=abs(I-V->D2);
  V->Dif[ 4]+=abs(I+V->D2);
  V->Dif[ 5]+=abs(I-V->D3);
  V->Dif[ 6]+=abs(I+V->D3);
  V->Dif[ 7]+=abs(I-V->D4);
  V->Dif[ 8]+=abs(I+V->D4);
  V->Dif[ 9]+=abs(I-ctx_rar->ChannelDelta);
  V->Dif[10]+=abs(I+ctx_rar->ChannelDelta);

  ctx_rar->ChannelDelta=V->LastDelta=(signed char)(Ch-V->LastChar);
  V->LastChar=Ch;

  if ((V->ByteCount & 0x1F)==0)
  {
    MinDif=V->Dif[0];
    NumMinDif=0;
    V->Dif[0]=0;
    for (I=1;(unsigned int)I<sizeof(V->Dif)/sizeof(V->Dif[0]);I++)
    {
      if (V->Dif[I]<MinDif)
      {
        MinDif=V->Dif[I];
        NumMinDif=I;
      }
      V->Dif[I]=0;
    }
    switch(NumMinDif)
    {
      case 1:
        if (V->K1>=-16)
          V->K1--;
        break;
      case 2:
        if (V->K1<16)
          V->K1++;
        break;
      case 3:
        if (V->K2>=-16)
          V->K2--;
        break;
      case 4:
        if (V->K2<16)
          V->K2++;
        break;
      case 5:
        if (V->K3>=-16)
          V->K3--;
        break;
      case 6:
        if (V->K3<16)
          V->K3++;
        break;
      case 7:
        if (V->K4>=-16)
          V->K4--;
        break;
      case 8:
        if (V->K4<16)
          V->K4++;
        break;
      case 9:
        if (V->K5>=-16)
          V->K5--;
        break;
      case 10:
        if (V->K5<16)
          V->K5++;
        break;
    }
  }
  return((UBYTE)Ch);
}}}
static void         DecodeNumber(struct sRAR *ctx_rar, struct Decode *Deco) {{{ // 52.6% of all CPU time is spent within this function!!!
  unsigned int I;
  register unsigned int N;
  GetBits();
#ifdef _USE_ASM

#ifdef _WIN_32
 __asm {

    xor eax, eax
    mov eax, ctx_rar->BitField                       /* N=ctx_rar->BitField & 0xFFFE; */
    and eax, 0xFFFFFFFE
    mov [N], eax
    mov edx, [Deco]                         /* EAX=N, EDX=Deco */

          cmp  eax, dword ptr[edx + 8*4 + 4] /* if (N<Dec->DecodeLen[8]) */
          jae  else_G

             cmp  eax, dword ptr[edx + 4*4 + 4] /* if (N<Dec->DecodeLen[4]) */
             jae  else_F


                cmp  eax, dword ptr[edx + 2*4 + 4] /* if (N<Dec->DecodeLen[2]) */
                jae  else_C

                   cmp  eax, dword ptr[edx + 1*4 + 4] /* if (N<Dec->DecodeLen[1]) */
                   jae  else_1
                   mov  I, 1                         /*  I=1; */
                   jmp  next_1
                 else_1:                             /* else */
                   mov  I, 2                         /*  I=2; */
                 next_1:

                jmp  next_C
              else_C:                             /* else */

                   cmp  eax, dword ptr[edx + 3*4 + 4] /* if (N<Dec->DecodeLen[3]) */
                   jae  else_2
                   mov  I, 3                         /*  I=3; */
                   jmp  next_2
                 else_2:                             /* else */
                   mov  I, 4                         /*  I=4; */
                 next_2:

              next_C:                             /* else */

             jmp  next_F
           else_F:


             cmp  eax, dword ptr[edx + 6*4 + 4] /* if (N<Dec->DecodeLen[6]) */
             jae  else_E

                cmp  eax, dword ptr[edx + 5*4 + 4] /* if (N<Dec->DecodeLen[5]) */
                jae  else_3
                mov  I, 5                         /*  I=5; */
                jmp  next_3
              else_3:                             /* else */
                mov  I, 6                         /*  I=6; */
              next_3:

             jmp  next_E
           else_E:                             /* else */

                cmp  eax, dword ptr[edx + 7*4 + 4] /* if (N<Dec->DecodeLen[7]) */
                jae  else_4
                mov  I, 7                         /*  I=7; */
                jmp  next_4
              else_4:                             /* else */
                mov  I, 8                         /*  I=8; */
              next_4:

           next_E:

           next_F:

          jmp  next_G
        else_G:

          cmp  eax, dword ptr[edx + 12*4 + 4] /* if (N<Dec->DecodeLen[12]) */
          jae  else_D

             cmp  eax, dword ptr[edx + 10*4 + 4] /* if (N<Dec->DecodeLen[10]) */
             jae  else_B

                cmp  eax, dword ptr[edx + 9*4 + 4] /* if (N<Dec->DecodeLen[9]) */
                jae  else_5
                mov  I, 9                         /*  I=9; */
                jmp  next_5
              else_5:                             /* else */
                mov  I, 10                         /*  I=10; */
              next_5:

             jmp  next_B
           else_B:                             /* else */

                cmp  eax, dword ptr[edx + 11*4 + 4] /* if (N<Dec->DecodeLen[11]) */
                jae  else_6
                mov  I, 11                         /*  I=11; */
                jmp  next_6
              else_6:                              /* else */
                mov  I, 12                         /*  I=12; */
              next_6:

           next_B:


          jmp  next_D
        else_D:                             /* else */

               cmp  eax, dword ptr[edx + 14*4 + 4] /* if (N<Dec->DecodeLen[14]) */
               jae  else_A

                  cmp  eax, dword ptr[edx + 13*4 + 4] /* if (N<Dec->DecodeLen[13]) */
                  jae  else_7
                  mov  I, 13                         /*  I=13; */
                  jmp  next_7
                 else_7:                             /* else */
                  mov  I, 14                         /*  I=14; */
                 next_7:

               jmp  next_A
              else_A:                             /* else */
               mov  I, 15                         /*  I=15; */
              next_A:

        next_D:
    next_G:
}
#else
 __asm__ __volatile__ (
     "andl $0xFFFFFFFE, %%eax"
"      movl %%eax, %1"
"          cmpl 8*4(%%edx), %%eax /* 5379 */"
"          jae  else_G"
""
"             cmpl 4*4(%%edx), %%eax"
"             jae  else_F"
""
"                cmpl 2*4(%%edx), %%eax"
"                jae  else_C"
""
"                   cmpl 1*4(%%edx), %%eax"
""
"                   jae  else_1"
"                   movl $1, %0"
"                   jmp  next_1"
"                 else_1:       "
"                   movl  $2, %0"
"                 next_1:"
"                "
"                jmp  next_C"
"              else_C:          "
""
"                   cmpl 3*4(%%edx), %%eax "
"                   jae  else_2"
"                   movl  $3, %0"
"                   jmp  next_2"
"                 else_2:       "
"                   movl  $4, %0"
"                 next_2:"
""
"              next_C:          "
""
"             jmp  next_F"
"           else_F:"
""
"             cmpl 6*4(%%edx), %%eax"
"             jae  else_E"
""
"                cmpl 5*4(%%edx), %%eax"
"                jae  else_3"
"                movl  $5, %0   "
"                jmp  next_3"
"              else_3:          "
"                movl  $6, %0   "
"              next_3:"
""
"             jmp  next_E"
"           else_E:             "
""
"                cmpl 7*4(%%edx), %%eax"
"                jae  else_4"
"                movl  $7, %0   "
"                jmp  next_4"
"              else_4:          "
"                movl  $8, %0   "
"              next_4:"
""
"           next_E:"
""
"           next_F:"
""
"          jmp  next_G"
"        else_G:"
""
"          cmpl 12*4(%%edx), %%eax"
"          jae  else_D"
""
"             cmpl 10*4(%%edx), %%eax"
"             jae  else_B"
""
"                cmpl 9*4(%%edx), %%eax"
"                jae  else_5"
"                movl  $9, %0   "
"                jmp  next_5"
"              else_5:          "
"                movl  $10, %0  "
"              next_5:"
""
"             jmp  next_B"
"           else_B:             "
""
"                cmpl 11*4(%%edx), %%eax"
" "
"                jae  else_6"
"                movl  $11, %0  "
"                jmp  next_6"
"              else_6:          "
"                movl  $12, %0  "
"              next_6:"
""
"           next_B:"
"      "
"        "
"          jmp  next_D"
"        else_D:                "
""
"               cmpl 14*4(%%edx), %%eax"
"               jae  else_A"
""
"                  cmpl 13*4(%%edx), %%eax"
"                  jae  else_7"
"                  movl  $13, %0"
"                  jmp  next_7"
"                 else_7:       "
"                  movl  $14, %0"
"                 next_7:"
""
"               jmp  next_A"
"              else_A:          "
"               movl  $15, %0   "
"              next_A:"
"          "
"        next_D:                             "
"    next_G:"
     : "=g" (I), "=r"(N)
     : "eax" ((long)ctx_rar->BitField), "edx"((long)Deco->DecodeLen)
      : "memory"
     );
#endif /* #ifdef _WIN_32 ... #elif defined _X86_ASM_ */

#else
  N=ctx_rar->BitField & 0xFFFE;
  if (N<Deco->DecodeLen[8])  {
    if (N<Deco->DecodeLen[4]) {
      if (N<Deco->DecodeLen[2]) {
        if (N<Deco->DecodeLen[1])
          I=1;
        else
          I=2;
      } else {
        if (N<Deco->DecodeLen[3])
          I=3;
        else
          I=4;
      }
    } else {
      if (N<Deco->DecodeLen[6])  {
        if (N<Deco->DecodeLen[5])
          I=5;
        else
          I=6;
      } else {
        if (N<Deco->DecodeLen[7])
          I=7;
        else
          I=8;
      }
   }
  } else {
    if (N<Deco->DecodeLen[12]) {
      if (N<Deco->DecodeLen[10]) {
        if (N<Deco->DecodeLen[9])
          I=9;
        else
          I=10;
      } else {
        if (N<Deco->DecodeLen[11])
          I=11;
        else
          I=12;
      }
    } else {
      if (N<Deco->DecodeLen[14]) {
        if (N<Deco->DecodeLen[13])
          I=13;
        else
          I=14;

      } else {
          I=15;
      }
    }

  }
#endif

  AddBits(I);
  if ((N=Deco->DecodePos[I]+((N-Deco->DecodeLen[I-1])>>(16-I)))>=Deco->MaxNum)
      N=0;
  ctx_rar->Number=Deco->DecodeNum[N];
}}}
static void         UnpReadBuf  (struct sRAR *ctx_rar, int FirstBuf) {{{
  int RetCode;
  if (FirstBuf)
  {
    ctx_rar->ReadTop=UnpRead(ctx_rar, ctx_rar->InBuf,sizeof(ctx_rar->InBuf));
    ctx_rar->InAddr=0;
  }
  else
  {
    memcpy(ctx_rar->InBuf,&ctx_rar->InBuf[sizeof(ctx_rar->InBuf)-32],32);
    ctx_rar->InAddr&=0x1f;
    RetCode=UnpRead(ctx_rar, &ctx_rar->InBuf[32],sizeof(ctx_rar->InBuf)-32);
    if (RetCode>0)
      ctx_rar->ReadTop=RetCode+32;
    else
      ctx_rar->ReadTop=ctx_rar->InAddr;
  }
}}}
static void         MakeDecodeTables(unsigned char *LenTab, struct Decode *Dec, int Size) {{{
  int LenCount[16],TmpPos[16],I;
  long M,N;
  memset(LenCount,0,sizeof(LenCount));
  for (I=0;I<Size;I++)
    LenCount[LenTab[I] & 0xF]++;

  LenCount[0]=0;
  for (TmpPos[0]=Dec->DecodePos[0]=Dec->DecodeLen[0]=0,N=0,I=1;I<16;I++)
  {
    N=2*(N+LenCount[I]);
    M=N<<(15-I);
    if (M>0xFFFF)
      M=0xFFFF;
    Dec->DecodeLen[I]=(unsigned int)M;
    TmpPos[I]=Dec->DecodePos[I]=Dec->DecodePos[I-1]+LenCount[I-1];
  }

  for (I=0;I<Size;I++)
    if (LenTab[I]!=0)
      Dec->DecodeNum[TmpPos[LenTab[I] & 0xF]++]=I;
  Dec->MaxNum=Size;
}}}
static void         ReadTables  (struct sRAR *ctx_rar) {{{
  UBYTE BitLength[BC];
  unsigned char Table[MC*4];
  int TableSize,N,I;
  if (ctx_rar->InAddr>sizeof(ctx_rar->InBuf)-25)
    UnpReadBuf(ctx_rar, 0);
  GetBits();
  ctx_rar->UnpAudioBlock=(ctx_rar->BitField & 0x8000);

  if (!(ctx_rar->BitField & 0x4000))
    memset(ctx_rar->UnpOldTable,0,sizeof(ctx_rar->UnpOldTable));
  AddBits(2);


  if (ctx_rar->UnpAudioBlock)
  {
    ctx_rar->UnpChannels=((ctx_rar->BitField>>12) & 3)+1;
    if (ctx_rar->CurChannel>=ctx_rar->UnpChannels)
      ctx_rar->CurChannel=0;
    AddBits(2);
    TableSize=MC*ctx_rar->UnpChannels;
  }
  else
    TableSize=NC+DC+RC;


  for (I=0;I<BC;I++)
  {
    GetBits();
    BitLength[I]=(UBYTE)(ctx_rar->BitField >> 12);
    AddBits(4);
  }
  MakeDecodeTables(BitLength,(struct Decode *)&ctx_rar->BD,BC);
  I=0;
  while (I<TableSize)
  {
    if (ctx_rar->InAddr>sizeof(ctx_rar->InBuf)-5)
      UnpReadBuf(ctx_rar, 0);
    DecodeNumber(ctx_rar, (struct Decode *)&ctx_rar->BD);
    if (ctx_rar->Number<16) {
      Table[I]=(ctx_rar->Number+ctx_rar->UnpOldTable[I]) & 0xf;
      I++;
    }
    else
      if (ctx_rar->Number==16)
      {
        GetBits();
        N=(ctx_rar->BitField >> 14)+3;
        AddBits(2);
        while (N-- > 0 && I<TableSize)
        {
          Table[I]=Table[I-1];
          I++;
        }
      }
      else
      {
        if (ctx_rar->Number==17)
        {
          GetBits();
          N=(ctx_rar->BitField >> 13)+3;
          AddBits(3);
        }
        else
        {
          GetBits();
          N=(ctx_rar->BitField >> 9)+11;
          AddBits(7);
        }
        while (N-- > 0 && I<TableSize)
          Table[I++]=0;
      }
  }
  if (ctx_rar->UnpAudioBlock)
    for (I=0;I<ctx_rar->UnpChannels;I++)
      MakeDecodeTables(&Table[I*MC],(struct Decode *)ctx_rar->MDPtr[I],MC);
  else
  {
    MakeDecodeTables(&Table[0],(struct Decode *)&ctx_rar->LD,NC);
    MakeDecodeTables(&Table[NC],(struct Decode *)&ctx_rar->DD,DC);
    MakeDecodeTables(&Table[NC+DC],(struct Decode *)&ctx_rar->RD,RC);
  }
  memcpy(ctx_rar->UnpOldTable,Table,sizeof(ctx_rar->UnpOldTable));
}}}
static void         ReadLastTables(struct sRAR *ctx_rar) {{{
  if (ctx_rar->ReadTop>=ctx_rar->InAddr+5)
  {
    if (ctx_rar->UnpAudioBlock)
    {
      DecodeNumber(ctx_rar, (struct Decode *)ctx_rar->MDPtr[ctx_rar->CurChannel]);
      if (ctx_rar->Number==256)
        ReadTables(ctx_rar);
    }
    else
    {
      DecodeNumber(ctx_rar, (struct Decode *)&ctx_rar->LD);
      if (ctx_rar->Number==269)
        ReadTables(ctx_rar);
    }
  }
}}}
static void         UnpInitData (struct sRAR *ctx_rar) {{{
  ctx_rar->InAddr=ctx_rar->InBit=0;
  
  if (!(ctx_rar->NewLhd.Flags & LHD_SOLID))
  {
    ctx_rar->ChannelDelta=ctx_rar->CurChannel=0;

#ifdef _USE_ASM

#ifdef _WIN_32                              /* Win32 with VisualC           */

    __asm {
        push edi
        push eax
        push ecx

        cld                                 /* increment EDI and ESI        */
        mov  al, 0x00
        mov  ecx, SIZE ctx_rar->AudV
        mov  edi, Offset ctx_rar->AudV
        rep  stosb                          /* clear memory                 */

        mov  ecx, SIZE ctx_rar->OldDist
        mov  edi, Offset ctx_rar->OldDist
        rep  stosb                          /* clear memory                 */

        mov  ecx, SIZE ctx_rar->UnpOldTable
        mov  edi, Offset ctx_rar->UnpOldTable
        rep  stosb                          /* clear memory                 */

        pop  ecx
        pop  eax
        pop  edi


        mov  [ctx_rar->OldDistPtr], 0
        mov  [ctx_rar->LastDist], 0
        mov  [ctx_rar->LastLength], 0
        mov  [ctx_rar->UnpPtr], 0
        mov  [ctx_rar->WrPtr], 0
        mov  [ctx_rar->OldDistPtr], 0
        mov  [ctx_rar->LastLength], 0
        mov  [ctx_rar->LastDist], 0
        mov  [ctx_rar->UnpPtr], 0
        mov  [ctx_rar->WrPtr], 0

    }
    memset(ctx_rar->UnpBuf,0,MAXWINSIZE);


#else                    /* unix/linux on i386 cpus */
    __asm__ __volatile (
"        cld                                 /* increment EDI and ESI        */"
"        movb $0x00, %%al"
"        movl %0, %%ecx"
"        movl %1, %%edi"
"        rep  "
"        stosb                              /* clear memory                 */"
""
"        movl %2, %%ecx"
"        mov  %3, %%edi"
"        rep  "
"        stosb                              /* clear memory                 */"
""
"        movl %4, %%ecx"
"        movl %5, %%edi"
"        rep  "
"        stosb                              /* clear memory                 */"
""
"        movl $0, (ctx_rar->OldDistPtr)"
"        movl $0, (ctx_rar->LastDist)"
"        movl $0, (ctx_rar->LastLength)"
"        movl $0, (ctx_rar->UnpPtr)"
"        movl $0, (ctx_rar->WrPtr)"
"        movl $0, (ctx_rar->OldDistPtr)"
"        movl $0, (ctx_rar->LastLength)"
"        movl $0, (ctx_rar->LastDist)"
"        movl $0, (ctx_rar->UnpPtr)"
"        movl $0, (ctx_rar->WrPtr)"
        :
        : "m" ((long)sizeof(ctx_rar->AudV)),
          "m" ((long)ctx_rar->AudV),
          "m" ((long)sizeof(ctx_rar->OldDist)),
          "m" ((long)ctx_rar->OldDist),
          "m" ((long)sizeof(ctx_rar->UnpOldTable)),
          "m" ((long)ctx_rar->UnpOldTable)
        : "memory", "edi", "eax", "ecx"
    );
    memset(ctx_rar->UnpBuf,0,MAXWINSIZE);
#endif

#else                                       /* unix/linux on non-i386 cpu  */
    memset(ctx_rar->AudV,0,sizeof(ctx_rar->AudV));
    memset(ctx_rar->OldDist,0,sizeof(ctx_rar->OldDist));
    ctx_rar->OldDistPtr=0;
    ctx_rar->LastDist=ctx_rar->LastLength=0;
    memset(ctx_rar->UnpBuf,0,MAXWINSIZE);
    memset(ctx_rar->UnpOldTable,0,sizeof(ctx_rar->UnpOldTable));
    ctx_rar->UnpPtr=ctx_rar->WrPtr=0;
#endif

  }
}}}
static void         Unpack      (struct sRAR *ctx_rar, unsigned char *UnpAddr, BOOL FileFound) {{{ // 38.3% of all CPU time is spent within this function!!!
  static const unsigned char LDecode[]={0,1,2,3,4,5,6,7,8,10,12,14,16,20,24,28,32,40,48,56,64,80,96,112,128,160,192,224};
  static const unsigned char LBits  []={0,0,0,0,0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4,5,5,5,5};
  static const int DDecode[]={0,1,2,3,4,6,8,12,16,24,32,48,64,96,128,192,256,384, 512,768,1024,1536,2048,3072,4096,6144,8192,12288,
                        16384,24576,32768U,49152U,65536,98304,131072,196608, 262144,327680,393216,458752,524288,589824,655360,
                        720896,786432,851968,917504,983040};
  static const unsigned char DBits[]=  {0,0,0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12,13,13,14,14,15,15,16,
                                  16,16,16,16,16,16,16,16,16,16,16,16,16};
  static const unsigned char SDDecode[]={0,4,8,16,32,64,128,192};
  static const unsigned char SDBits[]={2,2,3, 4, 5, 6,  6,  6};
  unsigned int Bits;

  ctx_rar->UnpBuf=UnpAddr;                           /* UnpAddr is a pointer to the  */
  UnpInitData(ctx_rar);                            /* unpack buffer                */
  UnpReadBuf(ctx_rar, 1);
  if (!(ctx_rar->NewLhd.Flags & LHD_SOLID)) ReadTables(ctx_rar);
   ctx_rar->DestUnpSize--;
  while (ctx_rar->DestUnpSize>=0) {
    ctx_rar->UnpPtr&=MAXWINMASK;
    if (ctx_rar->InAddr>sizeof(ctx_rar->InBuf)-30)
      UnpReadBuf(ctx_rar, 0);
    if (((ctx_rar->WrPtr-ctx_rar->UnpPtr) & MAXWINMASK)<270 && ctx_rar->WrPtr!=ctx_rar->UnpPtr) {
      if (FileFound) {
        if (ctx_rar->UnpPtr<ctx_rar->WrPtr) {
                        if((*ctx_rar->temp_output_buffer_offset + ((0-ctx_rar->WrPtr) & MAXWINMASK) + ctx_rar->UnpPtr) > ctx_rar->NewLhd.UnpSize)
                        {
                           debug_log("Fatal! Buffer overrun during decompression!");
                          ctx_rar->DestUnpSize=-1;

                    } 
		else {
              /* copy extracted data to output buffer                         */
              memcpy(ctx_rar->temp_output_buffer + *ctx_rar->temp_output_buffer_offset, &ctx_rar->UnpBuf[ctx_rar->WrPtr], (0-ctx_rar->WrPtr) & MAXWINMASK);
              /* update offset within buffer                                  */
              *ctx_rar->temp_output_buffer_offset+= (0-ctx_rar->WrPtr) & MAXWINMASK;
              /* copy extracted data to output buffer                         */
              memcpy(ctx_rar->temp_output_buffer + *ctx_rar->temp_output_buffer_offset, ctx_rar->UnpBuf, ctx_rar->UnpPtr);
              /* update offset within buffer                                  */
              *ctx_rar->temp_output_buffer_offset+=ctx_rar->UnpPtr;
                        }
        } else
        {
                        if((*ctx_rar->temp_output_buffer_offset + (ctx_rar->UnpPtr-ctx_rar->WrPtr)) > ctx_rar->NewLhd.UnpSize) {
                           	debug_log("Fatal! Buffer overrun during decompression!");
                          	ctx_rar->DestUnpSize=-1;
                    		} 
			else
                        {
                  /* copy extracted data to output buffer                       */
              memcpy(ctx_rar->temp_output_buffer + *ctx_rar->temp_output_buffer_offset,
                     &ctx_rar->UnpBuf[ctx_rar->WrPtr], ctx_rar->UnpPtr-ctx_rar->WrPtr);
              *ctx_rar->temp_output_buffer_offset+=ctx_rar->UnpPtr-ctx_rar->WrPtr;     /* update offset within buffer */
                    }

            }
      }
      ctx_rar->WrPtr=ctx_rar->UnpPtr;
    }

    if (ctx_rar->UnpAudioBlock) {
      DecodeNumber(ctx_rar, (struct Decode *)ctx_rar->MDPtr[ctx_rar->CurChannel]);
      if (ctx_rar->Number==256) {
        ReadTables(ctx_rar);
        continue;
      }
      ctx_rar->UnpBuf[ctx_rar->UnpPtr++]=DecodeAudio(ctx_rar, ctx_rar->Number);
      if (++ctx_rar->CurChannel==ctx_rar->UnpChannels)
        ctx_rar->CurChannel=0;
      ctx_rar->DestUnpSize--;
      continue;
    }

    DecodeNumber(ctx_rar, (struct Decode *)&ctx_rar->LD);
    if (ctx_rar->Number<256) {
      ctx_rar->UnpBuf[ctx_rar->UnpPtr++]=(UBYTE)ctx_rar->Number;
      ctx_rar->DestUnpSize--;
      continue;
    }
    if (ctx_rar->Number>269) {
      ctx_rar->Length=LDecode[ctx_rar->Number-=270]+3;
      if ((Bits=LBits[ctx_rar->Number])>0) {
        GetBits();
        ctx_rar->Length+=ctx_rar->BitField>>(16-Bits);
        AddBits(Bits);
      }

      DecodeNumber(ctx_rar, (struct Decode *)&ctx_rar->DD);
      ctx_rar->Distance=DDecode[ctx_rar->Number]+1;
      if ((Bits=DBits[ctx_rar->Number])>0) {
        GetBits();
        ctx_rar->Distance+=ctx_rar->BitField>>(16-Bits);
        AddBits(Bits);
      }

      if (ctx_rar->Distance>=0x40000L)
        ctx_rar->Length++;

      if (ctx_rar->Distance>=0x2000)
        ctx_rar->Length++;

       ctx_rar->LastDist=ctx_rar->OldDist[ctx_rar->OldDistPtr++ & 3]=ctx_rar->Distance;
       ctx_rar->DestUnpSize-=(ctx_rar->LastLength=ctx_rar->Length);
       while (ctx_rar->Length--)
       {
         ctx_rar->UnpBuf[ctx_rar->UnpPtr]=ctx_rar->UnpBuf[(ctx_rar->UnpPtr-ctx_rar->Distance) & MAXWINMASK];
         ctx_rar->UnpPtr=(ctx_rar->UnpPtr+1) & MAXWINMASK;
       }

      continue;
    }
    if (ctx_rar->Number==269)
    {
      ReadTables(ctx_rar);
      continue;
    }
    if (ctx_rar->Number==256)
    {
      ctx_rar->Length=ctx_rar->LastLength;
      ctx_rar->Distance=ctx_rar->LastDist;
       ctx_rar->LastDist=ctx_rar->OldDist[ctx_rar->OldDistPtr++ & 3]=ctx_rar->Distance;
       ctx_rar->DestUnpSize-=(ctx_rar->LastLength=ctx_rar->Length);
       while (ctx_rar->Length--)
       {
         ctx_rar->UnpBuf[ctx_rar->UnpPtr]=ctx_rar->UnpBuf[(ctx_rar->UnpPtr-ctx_rar->Distance) & MAXWINMASK];
         ctx_rar->UnpPtr=(ctx_rar->UnpPtr+1) & MAXWINMASK;
       }
      continue;
    }
    if (ctx_rar->Number<261)
    {
      ctx_rar->Distance=ctx_rar->OldDist[(ctx_rar->OldDistPtr-(ctx_rar->Number-256)) & 3];
      DecodeNumber(ctx_rar, (struct Decode *)&ctx_rar->RD);
      ctx_rar->Length=LDecode[ctx_rar->Number]+2;
      if ((Bits=LBits[ctx_rar->Number])>0)
      {
        GetBits();
        ctx_rar->Length+=ctx_rar->BitField>>(16-Bits);
        AddBits(Bits);
      }
      if (ctx_rar->Distance>=0x40000)
        ctx_rar->Length++;
      if (ctx_rar->Distance>=0x2000)
        ctx_rar->Length++;
      if (ctx_rar->Distance>=0x101)
        ctx_rar->Length++;
       ctx_rar->LastDist=ctx_rar->OldDist[ctx_rar->OldDistPtr++ & 3]=ctx_rar->Distance;
       ctx_rar->DestUnpSize-=(ctx_rar->LastLength=ctx_rar->Length);
       while (ctx_rar->Length--)
       {
         ctx_rar->UnpBuf[ctx_rar->UnpPtr]=ctx_rar->UnpBuf[(ctx_rar->UnpPtr-ctx_rar->Distance) & MAXWINMASK];
         ctx_rar->UnpPtr=(ctx_rar->UnpPtr+1) & MAXWINMASK;
       }
      continue;
    }
    if (ctx_rar->Number<270)
    {
      ctx_rar->Distance=SDDecode[ctx_rar->Number-=261]+1;
      if ((Bits=SDBits[ctx_rar->Number])>0)
      {
        GetBits();
        ctx_rar->Distance+=ctx_rar->BitField>>(16-Bits);
        AddBits(Bits);
      }
      ctx_rar->Length=2;
       ctx_rar->LastDist=ctx_rar->OldDist[ctx_rar->OldDistPtr++ & 3]=ctx_rar->Distance;
       ctx_rar->DestUnpSize-=(ctx_rar->LastLength=ctx_rar->Length);
       while (ctx_rar->Length--)
       {
         ctx_rar->UnpBuf[ctx_rar->UnpPtr]=ctx_rar->UnpBuf[(ctx_rar->UnpPtr-ctx_rar->Distance) & MAXWINMASK];
         ctx_rar->UnpPtr=(ctx_rar->UnpPtr+1) & MAXWINMASK;
       }
      continue;
   }
  }
  ReadLastTables(ctx_rar);

  if (FileFound)                            /* flush buffer                 */
  {

    if (ctx_rar->UnpPtr<ctx_rar->WrPtr)
    {
          if((*ctx_rar->temp_output_buffer_offset + ((0-ctx_rar->WrPtr) & MAXWINMASK) + ctx_rar->UnpPtr) > ctx_rar->NewLhd.UnpSize)
          {
            debug_log("Fatal! Buffer overrun during decompression!");
                ctx_rar->DestUnpSize=-1;
          } else
          {
        /* copy extracted data to output buffer                             */
        memcpy(ctx_rar->temp_output_buffer + *ctx_rar->temp_output_buffer_offset, &ctx_rar->UnpBuf[ctx_rar->WrPtr],
               (0-ctx_rar->WrPtr) & MAXWINMASK);
        /* update offset within buffer                                      */
        *ctx_rar->temp_output_buffer_offset+= (0-ctx_rar->WrPtr) & MAXWINMASK;
        /* copy extracted data to output buffer                             */
        memcpy(ctx_rar->temp_output_buffer + *ctx_rar->temp_output_buffer_offset, ctx_rar->UnpBuf, ctx_rar->UnpPtr);
        /* update offset within buffer                                      */
        *ctx_rar->temp_output_buffer_offset+=ctx_rar->UnpPtr;
          }
    } else
    {
          if((*ctx_rar->temp_output_buffer_offset + (ctx_rar->UnpPtr-ctx_rar->WrPtr)) > ctx_rar->NewLhd.UnpSize)
          {
                 debug_log("Fatal! Buffer overrun during decompression!");
                ctx_rar->DestUnpSize=-1;
          } else
          {
        /* copy extracted data to output buffer                             */
        memcpy(ctx_rar->temp_output_buffer + *ctx_rar->temp_output_buffer_offset, &ctx_rar->UnpBuf[ctx_rar->WrPtr],
               ctx_rar->UnpPtr-ctx_rar->WrPtr);
        /* update offset within buffer                                      */
        *ctx_rar->temp_output_buffer_offset+=ctx_rar->UnpPtr-ctx_rar->WrPtr;
          }
    }
  }

  ctx_rar->WrPtr=ctx_rar->UnpPtr;
}}}
static BOOL         ExtrFile    (struct sRAR *ctx_rar, int desc) {{{
  	BOOL ReturnCode=TRUE;
  	BOOL FileFound=FALSE;                     /* TRUE=use current extracted data FALSE=throw data away, wrong file */
  	int newdesc;
	ctx_rar->ArcPtr = NULL;
  	ctx_rar->MemRARFile->offset = 0;          /* start reading from offset 0  */
	if (ctx_rar->ArcPtr != NULL) {
		debug_log("Warning ctx_rar->ArcPtr != NULL\n");
		}
#ifndef  _USE_MEMORY_TO_MEMORY_DECOMPRESSION
  	/* open and identify archive                                              */
  	newdesc = dup(desc);
  	cli_dbgmsg("ExtrFile(): dup(%d) = %d\n", desc, newdesc);
  	if ((ctx_rar->ArcPtr=fdopen(newdesc,READBINARY))==NULL) {
    		debug_log("Error opening file.");
		ReturnCode = FALSE;
		goto ende;
		}
#endif
    	if (!IsArchive(ctx_rar)) {
      		debug_log("Not a RAR file");
		ReturnCode = FALSE;
		goto ende;
  		}
	if ((ctx_rar->UnpMemory=malloc(UNP_MEMORY))==NULL) {
    		cli_dbgmsg("unrarlib: Can't allocate memory for decompression!");
		ReturnCode = FALSE;
		goto ende;
  		}
	cli_dbgmsg("unrarlib: Allocated %d bytes.\n", UNP_MEMORY);
  	tseek(ctx_rar, ctx_rar->NewMhd.HeadSize-ctx_rar->MainHeadSize,SEEK_CUR);
  	/* do while file is not extracted and there's no error                    */
  	do {
    		if (ReadBlock(ctx_rar, FILE_HEAD | READSUBBLOCK) <= 0) { /* read name of the next file within the RAR archive  */
/*
 * 21.11.2000  UnQ  There's a problem with some linux distros when a file
 *                  can not be found in an archive.
 * 07.09.2004  ThL  Seems more like a logical bug in this lib, since it
 *		    appears to occur once for every archive.
 */
      			/* debug_log("Couldn't read next filename from archive (I/O error)."); */
      			ReturnCode=FALSE;
      			break;                                /* error, file not found in     */
			}                                       /* archive or I/O error         */
    		if (ctx_rar->BlockHead.HeadType==SUB_HEAD) {
      			debug_log("Sorry, sub-headers not supported.");
      			ReturnCode=FALSE;
      			break;                                /* error => exit                */
    			}

    if(TRUE == (FileFound=(stricomp(ctx_rar->ArgName, ctx_rar->ArcFileName) == 0)))
    /* *** file found! ***                                                  */
    {
      {
	cli_dbgmsg("unrarlib: Allocating %d bytes\n", ctx_rar->NewLhd.UnpSize);
        if((ctx_rar->temp_output_buffer=malloc(ctx_rar->NewLhd.UnpSize)) == NULL) { ;/* allocate memory for the*/
	    cli_errmsg("unrarlib: Can't malloc %d bytes\n", ctx_rar->NewLhd.UnpSize);
	    ReturnCode = FALSE;
	    break;
	}
      }
      *ctx_rar->temp_output_buffer_offset=0;         /* file. The default offset     */
                                            /* within the buffer is 0       */
    }

    /* in case of a solid archive, we need to decompress any single file till
     * we have found the one we are looking for. In case of normal archives
     * (recommended!!), we skip the files until we are sure that it is the
     * one we want.
     */
    if((ctx_rar->NewMhd.Flags & 0x08) || FileFound)
    {
      if (ctx_rar->NewLhd.UnpVer<13 || ctx_rar->NewLhd.UnpVer>UNP_VER)
      {
        cli_dbgmsg("unknown compression method: %d  (min=13 max=%d)\n", ctx_rar->NewLhd.UnpVer, UNP_VER);
        ReturnCode=FALSE;
        break;                              /* error, can't extract file!   */
      }

      ctx_rar->CurUnpRead=ctx_rar->CurUnpWrite=0;
      if ((*ctx_rar->Password!=0) && (ctx_rar->NewLhd.Flags & LHD_PASSWORD))
        	ctx_rar->Encryption=ctx_rar->NewLhd.UnpVer;
      else 	ctx_rar->Encryption=0;
      if (ctx_rar->Encryption) SetCryptKeys(ctx_rar, ctx_rar->Password);

      ctx_rar->UnpPackedSize=ctx_rar->NewLhd.PackSize;
      ctx_rar->DestUnpSize=ctx_rar->NewLhd.UnpSize;

      if (ctx_rar->NewLhd.Method==0x30) {
	cli_dbgmsg("unrarlib: Unstore method temporarily not supported\n");
        /* UnstoreFile(); */
        ReturnCode=FALSE;
        break;                              /* error, can't extract file! */
      } else
      {
	cli_dbgmsg("unrarlib: Unpack()\n");
        Unpack(ctx_rar, ctx_rar->UnpMemory, FileFound);
      }

#ifdef _DO_CRC32_CHECK                      /* calculate CRC32              */
      if((UBYTE*)ctx_rar->temp_output_buffer != NULL) {
        if(ctx_rar->NewLhd.FileCRC!=~CalcCRC32(0xFFFFFFFFL, (UBYTE*)ctx_rar->temp_output_buffer, ctx_rar->NewLhd.UnpSize)) {
          debug_log("CRC32 error - file couldn't be decompressed correctly!");
          ReturnCode=FALSE;
          break;                              /* error, can't extract file! */
        }
      }
#endif
    				}
    			tseek(ctx_rar, ctx_rar->NextBlockPos,SEEK_SET);
  		} while(stricomp(ctx_rar->ArgName, ctx_rar->ArcFileName) != 0);/* exit if file is extracted */
  	/* free memory, clear password and close archive                          */
  	if(ctx_rar->UnpMemory) free(ctx_rar->UnpMemory);
  	ctx_rar->UnpMemory=NULL;
ende:	if (ctx_rar->ArcPtr!=NULL){
      		fclose(ctx_rar->ArcPtr);
      		ctx_rar->ArcPtr = NULL;
      		lseek(desc, 0, SEEK_SET);
  		}
  	return ReturnCode;
	}}}

/* -- global functions ---------------------------------------------------- */

/* Get a file from a RAR file to the "output" buffer. The UniquE RAR FileLib
 * does everything from allocating memory, decrypting and unpacking the file
 * from the archive. TRUE is returned if the file could be successfully
 * extracted, else a FALSE indicates a failure.
 */
int urarlib_get_r (struct sRAR *ctx_rar, void *output, unsigned long *size, char *filename, int desc, char *libpassword) {{{
  BOOL  retcode = FALSE;
	RAR_init(ctx_rar);
#ifdef _DEBUG_LOG
  	int  str_offs;                            /* used for debug-strings       */
  	char DebugMsg[500];                       /* used to compose debug msg    */
  	if(debug_log_first_start) {
    		debug_log_first_start=FALSE;            /* only create a new log file   */
    		debug_init(_DEBUG_LOG_FILE);            /* on startup                   */
		}
#endif
  	InitCRC();                                /* init some vars               */
  	strcpy(ctx_rar->ArgName, filename);                /* set file(s) to extract       */
#ifdef _USE_MEMORY_TO_MEMORY_DECOMPRESSION
  	ctx_rar->MemRARFile = rarfile;                     /* set pointer to mem-RAR file  */
#endif
  	if(libpassword != NULL) strcpy(ctx_rar->Password, libpassword);          /* init password                */
  	ctx_rar->temp_output_buffer = NULL;
  	ctx_rar->temp_output_buffer_offset=size;           /* set size of the temp buffer  */
  	retcode = ExtrFile(ctx_rar, desc);                     /* unpack file now! */
  	memset(ctx_rar->Password,0,sizeof(ctx_rar->Password));      /* clear password               */
#ifndef _USE_MEMORY_TO_MEMORY_DECOMPRESSION
  	if (ctx_rar->ArcPtr!=NULL){
      		/* FIXME: possible FILE* leak */
      		cli_dbgmsg("%s:%d NOT Close ctx_rar->ArcPtr from fd %d\n", __FILE__, __LINE__, desc);
      		/*
      		fclose(ctx_rar->ArcPtr);
      		lseek(desc, 0, SEEK_SET);
      		ctx_rar->ArcPtr = NULL;
      		*/
  		}
#endif
  	if(ctx_rar->UnpMemory ) free(ctx_rar->UnpMemory );
  	if(ctx_rar->TempMemory) free(ctx_rar->TempMemory);
  	if(ctx_rar->CommMemory) free(ctx_rar->CommMemory);
  	ctx_rar->UnpMemory =NULL;
  	ctx_rar->TempMemory=NULL;
  	ctx_rar->CommMemory=NULL;
  	if(retcode == FALSE) {
    		if(ctx_rar->temp_output_buffer) free(ctx_rar->temp_output_buffer);     /* free memory and return NULL  */
    		ctx_rar->temp_output_buffer=NULL;
    		*(DWORD*)output=0;                      /* pointer on errors            */
    		*size=0;
#ifdef _DEBUG_LOG
   		/* sorry for this ugly code, but older SunOS gcc compilers don't support */
   		/* white spaces within strings                                           */
   		str_offs  = sprintf(DebugMsg, "Error - couldn't extract ");
   		str_offs += sprintf(DebugMsg + str_offs, ">%s<", filename);
   		str_offs += sprintf(DebugMsg + str_offs, " and allocated ");
   		str_offs += sprintf(DebugMsg + str_offs, "%u Bytes", (unsigned int)*size);
   		str_offs += sprintf(DebugMsg + str_offs, " of unused memory!");
	  	debug_log(DebugMsg);
  		} 
  	else {
    		sprintf(DebugMsg, "Extracted %u Bytes.", (unsigned int)*size);
	  	debug_log(DebugMsg);
#endif
		}
  	*(DWORD*)output=(DWORD)ctx_rar->temp_output_buffer;/* return pointer for unpacked data*/
  	return retcode;
	}}}
int urarlib_get   (                      void *output, unsigned long *size, char *filename, int desc, char *libpassword) {{{
	struct sRAR *ctx_rar = &ctx_rar_buf;
	ctx_rar->is_init = 0;
	return urarlib_get_r(ctx_rar, output, size, filename, desc, libpassword);
	}}}

int urarlib_list_r(struct sRAR *ctx_rar, int desc, ArchiveList_struct *list) {{{
  ArchiveList_struct *tmp_List = NULL;
  int NoOfFilesInArchive       = 0;         		/* number of files in archive   */
  int newdesc;
	RAR_init(ctx_rar);

#ifdef _DEBUG_LOG
  if(debug_log_first_start) {
    	debug_log_first_start=FALSE;            	/* only create a new log file   */
    	debug_init(_DEBUG_LOG_FILE);            	/* on startup                   */
	}
#endif
  	InitCRC();                             		/* init some vars               */
	ctx_rar->ArcPtr = NULL;
#ifdef _USE_MEMORY_TO_MEMORY_DECOMPRESSION
  	ctx_rar->MemRARFile         = rarfile;          /* assign pointer to RAR file   */
  	ctx_rar->MemRARFile->offset = 0;
#else
  	/* open and identify archive                                              */
  	newdesc = dup(desc);
  	cli_dbgmsg("ExtrFile(): dup(%d) = %d\n", desc, newdesc);
  	if ((ctx_rar->ArcPtr=fdopen(newdesc,READBINARY))==NULL) {
    		cli_dbgmsg("urarlib_list(): Error opening file: %s", strerror(errno));
    		debug_log("Error opening file.");
    		cli_dbgmsg("%s:%d Close fd %d\n", __FILE__, __LINE__, newdesc);
    		close(newdesc);
    		return NoOfFilesInArchive;
		}
#endif
  	if (!IsArchive(ctx_rar)) {
    		debug_log("Not a RAR file");
		cli_dbgmsg("urarlib_list(): Not a valid archive.");
		goto ende;
		}
  	if ((ctx_rar->UnpMemory=malloc(UNP_MEMORY))==NULL) {
    		cli_dbgmsg("urarlib_list(): out of memory.");
    		debug_log("Can't allocate memory for decompression!");
		goto ende;
		}
  tseek(ctx_rar, ctx_rar->NewMhd.HeadSize-ctx_rar->MainHeadSize,SEEK_CUR);
  (*(DWORD*)list) = (DWORD)NULL;            /* init file list               */
  /* do while file is not extracted and there's no error                    */
  while (TRUE) {
    int ReadBlockResult;
    if ((ReadBlockResult = ReadBlock(ctx_rar, FILE_HEAD | READSUBBLOCK)) <= 0) /* read name of the next  */
    {                                       /* file within the RAR archive  */
      cli_dbgmsg("Couldn't read next filename from archive (I/O error): %d\n", ReadBlockResult);
      break;                                /* error, file not found in     */
    }                                       /* archive or I/O error         */
    		if (ctx_rar->BlockHead.HeadType==SUB_HEAD) {
      			debug_log("Sorry, sub-headers not supported.");
      			NoOfFilesInArchive = 0;
      			break;                           /* error => exit                */
    			}
    		if((void*)(*(DWORD*)list) == NULL) {     /* first entry                  */
      			tmp_List = malloc(sizeof(ArchiveList_struct));	// TODO
      			tmp_List->next = NULL;
      			(*(DWORD*)list) = (DWORD)tmp_List;
			}
     		else {                                 /* add entry                    */
      			tmp_List->next = malloc(sizeof(ArchiveList_struct));
      			tmp_List = (ArchiveList_struct*) tmp_List->next;
      			tmp_List->next = NULL;
    			}
		tmp_List->item.Name = malloc(ctx_rar->NewLhd.NameSize + 1);
		strcpy(tmp_List->item.Name, ctx_rar->ArcFileName);
		tmp_List->item.NameSize	= ctx_rar->NewLhd.NameSize;
		tmp_List->item.PackSize	= ctx_rar->NewLhd.PackSize;
		tmp_List->item.UnpSize 	= ctx_rar->NewLhd.UnpSize;
		tmp_List->item.HostOS 	= ctx_rar->NewLhd.HostOS;
		tmp_List->item.FileCRC 	= ctx_rar->NewLhd.FileCRC;
		tmp_List->item.FileTime	= ctx_rar->NewLhd.FileTime;
		tmp_List->item.UnpVer 	= ctx_rar->NewLhd.UnpVer;
		tmp_List->item.Method 	= ctx_rar->NewLhd.Method;
		tmp_List->item.FileAttr	= ctx_rar->NewLhd.FileAttr;
		tmp_List->item.Flags 	= ctx_rar->NewLhd.Flags;
		NoOfFilesInArchive++;                   /* count files                  */
    		tseek(ctx_rar, ctx_rar->NextBlockPos,SEEK_SET);
  		} // while (TRUE)
  	/* free memory, clear password and close archive                          */
  	memset(ctx_rar->Password,0,sizeof(ctx_rar->Password));      /* clear password               */
  	if(ctx_rar->UnpMemory ) free(ctx_rar->UnpMemory );
  	if(ctx_rar->TempMemory) free(ctx_rar->TempMemory);
  	if(ctx_rar->CommMemory) free(ctx_rar->CommMemory);
  	ctx_rar->UnpMemory =NULL;
  	ctx_rar->TempMemory=NULL;
  	ctx_rar->CommMemory=NULL;

ende:	if (ctx_rar->ArcPtr!=NULL){
      		fclose(ctx_rar->ArcPtr);
      		ctx_rar->ArcPtr = NULL;
      		lseek(desc, 0, SEEK_SET);
		}
  	return NoOfFilesInArchive;
	}}}
int urarlib_list  (                      int desc, ArchiveList_struct *list) {{{
	struct sRAR *ctx_rar = &ctx_rar_buf;
	ctx_rar->is_init = 0;
	return urarlib_list_r(ctx_rar, desc, list);
	}}}

/* urarlib_freelist:
 * (after the suggestion and code of Duy Nguyen, Sean O'Blarney
 * and Johannes Winkelmann who independently wrote a patch)
 * free the memory of a ArchiveList_struct created by urarlib_list.
 *
 *    input: *list          pointer to an ArchiveList_struct
 *    output: -
 */
void urarlib_freelist(ArchiveList_struct *list) {{{
    	ArchiveList_struct* tmp = list;
    	while ( list ) {
        	tmp = list->next;
        	free( list->item.Name );
        	free( list );
        	list = tmp;
    		}
	}}}

/* ****************************************
 * ** RAR decompression code starts here **
 * ****************************************/

void SetOldKeys(char *NewPassword) {{{
  unsigned char PN1,PN2,PN3;				
  unsigned short OldKey[4];				
  UDWORD PswCRC;					
  UBYTE Ch;						
  PswCRC=CalcCRC32(0xFFFFFFFFL,(UBYTE*)NewPassword,strlen(NewPassword));
  OldKey[0]=(UWORD)PswCRC;				
  OldKey[1]=(UWORD)(PswCRC>>16);			
  OldKey[2]=0;						
  OldKey[3]=0;						
  PN1=PN2=PN3=0;					
  while ((Ch=*NewPassword)!=0)				
  {							
    PN1+=Ch;						
    PN2^=Ch;						
    PN3+=Ch;						
    PN3=(UBYTE)rol(PN3,1);				
    OldKey[2]^=((UWORD)(Ch^CRCTab[Ch]));		
    OldKey[3]+=((UWORD)(Ch+(CRCTab[Ch]>>16)));		
    NewPassword++;					
  }							
}}}

#ifdef _DEBUG_LOG /* Debug Functions {{{ */

#ifdef _UNIX
DWORD GetTickCount() {
    	struct timeval tv;
    	gettimeofday( &tv, 0 );
    	return (tv.tv_usec / 1000);
	}
#endif

/* -- global stuff -------------------------------------------------------- */
char  log_file_name[256];                   /* file name for the log file   */
DWORD debug_start_time;                     /* starttime of debug           */
BOOL  debug_started = FALSE;                /* debug_log writes only if     */
                                            /* this is TRUE                 */
/* ------------------------------------------------------------------------ */

/* -- global functions ---------------------------------------------------- */
void debug_init_proc(char *file_name)
/* Create/Rewrite a log file                                                */
{{{
  FILE *fp;
  char date[] = __DATE__;
  char time[] = __TIME__;

  debug_start_time = GetTickCount();        /* get start time               */
  strcpy(log_file_name, file_name);         /* save file name               */

  if((fp = fopen(log_file_name, CREATETEXT)) != NULL)
  {
    debug_started = TRUE;                   /* enable debug                 */
    fprintf(fp, "Debug log of UniquE's RARFileLib\n"\
                "~^~~~~~~~~~~~~~~~~~~~~~~~~~~~~^~\n");
    fprintf(fp, "(executable compiled on %s at %s)\n\n", date, time);
    fclose(fp);
  }
}}}


void debug_log_proc(char *text, char *sourcefile, int sourceline)
/* add a line to the log file                                               */
{{{
  FILE *fp;

  if(debug_started == FALSE) return;        /* exit if not initialized      */

  if((fp = fopen(log_file_name, APPENDTEXT)) != NULL) /* append to logfile  */

  {
    fprintf(fp, " %8u ms (line %u in %s):\n              - %s\n",
            (unsigned int)(GetTickCount() - debug_start_time),
            sourceline, sourcefile, text);
    fclose(fp);
  }
}}}

/* ------------------------------------------------------------------------ */
#endif	/* }}} */

/* end of file urarlib.c */
